__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import atexit
import logging
import os
import subprocess
import threading
import time
from http import HTTPStatus
from typing import Any, Dict, List, Optional
from urllib.request import Request, urlopen

import yaml

from pyngrok import conf, installer
from pyngrok.conf import PyngrokConfig
from pyngrok.exception import PyngrokError, PyngrokNgrokError, PyngrokSecurityError
from pyngrok.installer import SUPPORTED_NGROK_VERSIONS
from pyngrok.log import NgrokLog

logger = logging.getLogger(__name__)
ngrok_logger = logging.getLogger(f"{__name__}.ngrok")


class NgrokProcess:
    """
    An object containing information about the ``ngrok`` process.
    """

    def __init__(self,
                 proc: subprocess.Popen,  # type: ignore
                 pyngrok_config: PyngrokConfig) -> None:
        #: The child process that is running ``ngrok``.
        self.proc: subprocess.Popen = proc  # type: ignore
        #: The ``pyngrok`` configuration to use with ``ngrok``.
        self.pyngrok_config: PyngrokConfig = pyngrok_config

        #: The API URL for the ``ngrok`` web interface.
        self.api_url: Optional[str] = None
        #: A list of the most recent logs from ``ngrok``, limited in size to ``max_logs``.
        self.logs: List[NgrokLog] = []
        #: If ``ngrok`` startup fails, this will be the log of the failure.
        self.startup_error: Optional[str] = None

        self._tunnel_started = False
        self._client_connected = False
        self._monitor_thread: Optional[threading.Thread] = None
        self._monitor_thread_alive = False

    def __repr__(self) -> str:
        return f"<NgrokProcess: \"{self.api_url}\">"

    def __str__(self) -> str:  # pragma: no cover
        return f"NgrokProcess: \"{self.api_url}\""

    @staticmethod
    def _line_has_error(log: NgrokLog) -> bool:
        return log.lvl in ["ERROR", "CRITICAL"]

    def _log_startup_line(self, line: str) -> Optional[NgrokLog]:
        """
        Parse the given startup log line and use it to manage the startup state
        of the ``ngrok`` process.

        :param line: The line to be parsed and logged.
        :return: The parsed log.
        """
        log = self._log_line(line)

        if log is None:
            return None
        elif self._line_has_error(log):
            self.startup_error = log.err
        elif log.msg:
            # Log ngrok startup states as they come in
            if "starting web service" in log.msg and log.addr is not None:
                self.api_url = f"http://{log.addr}"
            elif "tunnel session started" in log.msg:
                self._tunnel_started = True
            elif "client session established" in log.msg:
                self._client_connected = True

        return log

    def _log_line(self, line: str) -> Optional[NgrokLog]:
        """
        Parse, log, and emit (if ``log_event_callback`` in :class:`~pyngrok.conf.PyngrokConfig` is registered) the
        given log line.

        :param line: The line to be processed.
        :return: The parsed log.
        """
        log = NgrokLog(line)

        if log.line == "":
            return None

        lvl = log.lvl if log.lvl != "NOTSET" else "INFO"

        ngrok_logger.log(getattr(logging, lvl), log.line)
        self.logs.append(log)
        if len(self.logs) > self.pyngrok_config.max_logs:
            self.logs.pop(0)

        if self.pyngrok_config.log_event_callback is not None:
            self.pyngrok_config.log_event_callback(log)

        return log

    def healthy(self) -> bool:
        """
        Check whether the ``ngrok`` process has finished starting up and is in a running, healthy state.

        :return: ``True`` if the ``ngrok`` process is started, running, and healthy.
        :raises: :class:`~pyngrok.exception.PyngrokSecurityError`: When the ``url`` is not supported.
        """
        if self.api_url is None or \
                not self._tunnel_started or \
                not self._client_connected:
            return False

        if not self.api_url.lower().startswith("http"):
            raise PyngrokSecurityError(f"URL must start with \"http\": {self.api_url}")

        # Ensure the process is available for requests before registering it as healthy
        request = Request(f"{self.api_url}/api/tunnels")
        response = urlopen(request)
        if response.getcode() != HTTPStatus.OK:
            return False

        return self.proc.poll() is None

    def _monitor_process(self) -> None:
        self._monitor_thread_alive = True
        while self._monitor_thread_alive and self.proc.poll() is None:
            if self.proc.stdout is None:
                logger.debug("Output from process is empty, nothing to log")

                continue

            self._log_line(self.proc.stdout.readline())

        self._monitor_thread = None

    def start_monitor_thread(self) -> None:
        """
        Start a thread that will monitor the ``ngrok`` process and its logs until it completes.

        If a monitor thread is already running, nothing will be done.
        """
        if self._monitor_thread is None:
            logger.debug("Monitor thread will be started")

            self._monitor_thread = threading.Thread(target=self._monitor_process,
                                                    daemon=True)
            self._monitor_thread.start()

    def stop_monitor_thread(self) -> None:
        """
        Set the monitor thread to stop monitoring the ``ngrok`` process after the next log event. This will not
        necessarily terminate the thread immediately, as the thread may currently be idle, rather it sets a flag
        on the thread telling it to terminate the next time it wakes up.

        This has no impact on the ``ngrok`` process itself, only ``pyngrok``'s monitor of the process and
        its logs.
        """
        if self._monitor_thread is not None:
            logger.debug("Monitor thread will be stopped")

            self._monitor_thread_alive = False


def set_auth_token(pyngrok_config: PyngrokConfig,
                   token: str) -> None:
    """
    Set the ``ngrok`` auth token in the config file to streamline access to more features (for instance, multiple
    concurrent tunnels, custom domains, etc.).

    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary.
    :param token: The auth token to set.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the ``ngrok_version`` is not supported.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokError`: When ``ngrok`` could not start.
    """
    if pyngrok_config.ngrok_version == "v2":
        start = [pyngrok_config.ngrok_path, "authtoken", token, "--log", "stdout"]
    elif pyngrok_config.ngrok_version == "v3":
        start = [pyngrok_config.ngrok_path, "config", "add-authtoken", token, "--log", "stdout"]
    else:
        raise PyngrokError(f"\"ngrok_version\" must be a supported version: {SUPPORTED_NGROK_VERSIONS}")

    if pyngrok_config.config_path:
        logger.info(f"Updating authtoken for \"config_path\": {pyngrok_config.config_path}")
        start.append("--config")
        start.append(pyngrok_config.config_path)
    else:
        logger.info(
            f"Updating authtoken for default \"config_path\" of \"ngrok_path\": {pyngrok_config.ngrok_path}")

    result = str(subprocess.check_output(start))

    if "Authtoken saved" not in result:
        raise PyngrokNgrokError(f"An error occurred when saving the auth token: {result}")


def set_api_key(pyngrok_config: PyngrokConfig,
                key: str) -> None:
    """
    Set the ``ngrok`` API key in the config file to enable access to more features (for instance,
    `Internal Endpoints <https://ngrok.com/docs/universal-gateway/internal-endpoints/>`_).

    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary.
    :param key: The API key to set.
    :raises: :class:`~pyngrok.exception.PyngrokError`: When the ``ngrok_version`` is not supported.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokError`: When ``ngrok`` could not start.
    """
    if pyngrok_config.ngrok_version == "v2":
        raise PyngrokError("\"ngrok_version\" v2 does not have this command.")
    elif pyngrok_config.ngrok_version == "v3":
        start = [pyngrok_config.ngrok_path, "config", "add-api-key", key, "--log", "stdout"]
    else:
        raise PyngrokError(f"\"ngrok_version\" must be a supported version: {SUPPORTED_NGROK_VERSIONS}")

    if pyngrok_config.config_path:
        logger.info(f"Updating API key for \"config_path\": {pyngrok_config.config_path}")
        start.append("--config")
        start.append(pyngrok_config.config_path)
    else:
        logger.info(
            f"Updating API key for default \"config_path\" of \"ngrok_path\": {pyngrok_config.ngrok_path}")

    result = str(subprocess.check_output(start))

    if "API key saved" not in result:
        raise PyngrokNgrokError(f"An error occurred when saving the API key: {result}")


def is_process_running(ngrok_path: str) -> bool:
    """
    Check if the ``ngrok`` process is currently running.

    :param ngrok_path: The path to the ``ngrok`` binary.
    :return: ``True`` if ``ngrok`` is running from the given path.
    """
    if ngrok_path in _current_processes:
        # Ensure the process is still running and hasn't been killed externally, otherwise cleanup
        if _current_processes[ngrok_path].proc.poll() is None:
            return True
        else:
            logger.debug(
                f"Removing stale process for \"ngrok_path\" {ngrok_path}")

            _current_processes.pop(ngrok_path, None)

    return False


def get_process(pyngrok_config: PyngrokConfig) -> NgrokProcess:
    """
    Get the current ``ngrok`` process for the given config's ``ngrok_path``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary.
    :return: The ``ngrok`` process.
    """
    if is_process_running(pyngrok_config.ngrok_path):
        return _current_processes[pyngrok_config.ngrok_path]

    return _start_process(pyngrok_config)


def kill_process(ngrok_path: str) -> None:
    """
    Terminate the ``ngrok`` processes, if running, for the given path. This method will not block, it will just
    issue a kill request.

    :param ngrok_path: The path to the ``ngrok`` binary.
    """
    if is_process_running(ngrok_path):
        ngrok_process = _current_processes[ngrok_path]

        logger.info(f"Killing ngrok process: {ngrok_process.proc.pid}")

        try:
            ngrok_process.proc.kill()
            ngrok_process.proc.wait()
        except OSError as e:  # pragma: no cover
            # If the process was already killed, nothing to do but cleanup state
            if e.errno != 3:
                raise e

        _current_processes.pop(ngrok_path, None)
    else:
        logger.debug(f"\"ngrok_path\" {ngrok_path} is not running a process")


def run_process(ngrok_path: str, args: List[str]) -> None:
    """
    Start a blocking ``ngrok`` process with the binary at the given path and the passed args.

    This method is meant for invoking ``ngrok`` directly (for instance, from the command line) and is not
    necessarily compatible with non-blocking API methods. For that, use :func:`~pyngrok.process.get_process`.

    :param ngrok_path: The path to the ``ngrok`` binary.
    :param args: The args to pass to ``ngrok``.
    """
    _validate_path(ngrok_path)

    start = [ngrok_path] + args
    subprocess.call(start)


def capture_run_process(ngrok_path: str, args: List[str]) -> str:
    """
    Start a blocking ``ngrok`` process with the binary at the given path and the passed args. When the process
    returns, so will this method, and the captured output from the process along with it.

    This method is meant for invoking ``ngrok`` directly (for instance, for API requests) and is not
    necessarily compatible with non-blocking API methods or interacting with active tunnels. For that, use
    :func:`~pyngrok.process.get_process`.

    :param ngrok_path: The path to the ``ngrok`` binary.
    :param args: The args to pass to ``ngrok``.
    :return: The output from the process.
    :raises: PyngrokNgrokError The ``ngrok`` process exited with an error.
    :raises: CalledProcessError An error occurred while executing the process.
    """
    _validate_path(ngrok_path)

    start = [ngrok_path] + args
    try:
        output = subprocess.check_output(start, stderr=subprocess.STDOUT)

        return output.decode("utf-8").strip()
    except subprocess.CalledProcessError as e:
        if e.returncode != 0:
            raise PyngrokNgrokError(f"The ngrok process exited with code {e.returncode}: "
                                    f"{e.output.decode('utf-8').strip()}")
        else:
            raise e


def _validate_path(ngrok_path: str) -> None:
    """
    Validate the given path exists, is a ``ngrok`` binary, and is ready to be started, otherwise raise a
    relevant exception.

    :param ngrok_path: The path to the ``ngrok`` binary.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokError`: When ``ngrok`` could not start.
    """
    if not os.path.exists(ngrok_path):
        raise PyngrokNgrokError(
            f"ngrok binary was not found. Be sure to call \"ngrok.install_ngrok()\" first "
            f"for \"ngrok_path\": {ngrok_path}")

    if ngrok_path in _current_processes:
        raise PyngrokNgrokError(f"ngrok is already running for the \"ngrok_path\": {ngrok_path}")


def _validate_config(config_path: str) -> None:
    with installer.config_file_lock:
        with open(config_path, "r") as config_file:
            config = yaml.safe_load(config_file)

    if config is not None:
        installer.validate_config(config)


def _terminate_process(process: subprocess.Popen) -> None:  # type: ignore
    if process is None:
        return

    try:
        process.terminate()
    except OSError:  # pragma: no cover
        logger.debug(f"ngrok process already terminated: {process.pid}")


def _start_process(pyngrok_config: PyngrokConfig) -> NgrokProcess:
    """
    Start a ``ngrok`` process with no tunnels. This will start the ``ngrok`` web interface, against
    which HTTP requests can be made to create, interact with, and destroy tunnels.

    :param pyngrok_config: The ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary.
    :return: The ``ngrok`` process.
    :raises: :class:`~pyngrok.exception.PyngrokNgrokError`: When ``ngrok`` could not start.
    """
    config_path = conf.get_config_path(pyngrok_config)

    _validate_path(pyngrok_config.ngrok_path)
    _validate_config(config_path)

    start = [pyngrok_config.ngrok_path, "start", "--none", "--log", "stdout"]
    if pyngrok_config.config_path:
        logger.info(f"Starting ngrok with config file: {pyngrok_config.config_path}")
        start.append("--config")
        start.append(pyngrok_config.config_path)
    if pyngrok_config.auth_token:
        logger.info("Overriding default auth token")
        start.append("--authtoken")
        start.append(pyngrok_config.auth_token)
    if pyngrok_config.region:
        logger.info(f"Starting ngrok in region: {pyngrok_config.region}")
        start.append("--region")
        start.append(pyngrok_config.region)

    popen_kwargs: Dict[str, Any] = {"stdout": subprocess.PIPE, "universal_newlines": True}
    if os.name == "posix":
        popen_kwargs.update(start_new_session=pyngrok_config.start_new_session)
    elif pyngrok_config.start_new_session:
        logger.warning("Ignoring start_new_session=True, which requires POSIX")
    proc = subprocess.Popen(start, **popen_kwargs)
    atexit.register(_terminate_process, proc)

    logger.debug(f"ngrok process starting with PID: {proc.pid}")

    ngrok_process = NgrokProcess(proc, pyngrok_config)
    _current_processes[pyngrok_config.ngrok_path] = ngrok_process

    timeout = time.time() + pyngrok_config.startup_timeout
    while time.time() < timeout:
        if proc.stdout is None:
            logger.debug("Output from process is empty, breaking startup loop")
            break

        line = proc.stdout.readline()
        ngrok_process._log_startup_line(line)

        if ngrok_process.healthy():
            logger.debug(f"ngrok process has started with API URL: {ngrok_process.api_url}")

            ngrok_process.startup_error = None

            if pyngrok_config.monitor_thread:
                ngrok_process.start_monitor_thread()

        if ngrok_process.healthy() or ngrok_process.proc.poll() is not None:
            break

    if not ngrok_process.healthy():
        # If the process did not come up in a healthy state, clean up the state
        kill_process(pyngrok_config.ngrok_path)

        if ngrok_process.startup_error is not None:
            raise PyngrokNgrokError(f"The ngrok process errored on start: {ngrok_process.startup_error}.",
                                    ngrok_process.logs,
                                    ngrok_process.startup_error)
        else:
            raise PyngrokNgrokError("The ngrok process was unable to start.", ngrok_process.logs)

    return ngrok_process


_current_processes: Dict[str, NgrokProcess] = {}



__copyright__ = "Copyright (c) 2018-2024 Alex Laird"
__license__ = "MIT"

import logging
import shlex
from typing import Optional


class NgrokLog:
    """
    An object containing a parsed log from the ``ngrok`` process.
    """

    def __init__(self,
                 line: str) -> None:
        #: The raw, unparsed log line.
        self.line: str = line.strip()

        #: The log's ISO 8601 timestamp.
        self.t: Optional[str] = None
        #: The log's level.
        self.lvl: str = "NOTSET"
        #: The log's message.
        self.msg: Optional[str] = None
        #: The log's error, if applicable.
        self.err: Optional[str] = None
        #: The log's type.
        self.obj: Optional[str] = None
        #: The URL, if ``obj`` is "web".
        self.addr: Optional[str] = None

        try:
            split = shlex.split(self.line)
        except ValueError:
            split = []

        for i in split:
            if "=" not in i:
                continue

            key, value = i.split("=", 1)

            if key == "lvl":
                if not value:
                    value = self.lvl

                value = value.upper()
                if value == "CRIT":
                    value = "CRITICAL"
                elif value in ["ERR", "EROR"]:
                    value = "ERROR"
                elif value == "WARN":
                    value = "WARNING"

                if not hasattr(logging, value):
                    value = self.lvl

            setattr(self, key, value)

    def __repr__(self) -> str:
        return f"<NgrokLog: t={self.t} lvl={self.lvl} msg=\"{self.msg}\">"

    def __str__(self) -> str:  # pragma: no cover
        attrs = [attr for attr in dir(self) if not attr.startswith("_") and getattr(self, attr) is not None]
        attrs.remove("line")

        return " ".join(f"{attr}=\"{getattr(self, attr)}\"" for attr in attrs)



__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import os
import time
import unittest
from urllib.parse import urlparse
from urllib.request import urlopen

from pyngrok import ngrok, agent
from tests.testcase import NgrokTestCase


class TestAgent(NgrokTestCase):
    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_captured_requests(self):
        # GIVEN
        tunnel_name = "my-tunnel"
        current_process = ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        public_url = ngrok.connect(urlparse(current_process.api_url).port, name=tunnel_name,
                                   pyngrok_config=self.pyngrok_config_v3).public_url
        time.sleep(1)

        urlopen(f"{public_url}/api/status").read()
        time.sleep(3)

        # WHEN
        response1 = agent.get_requests(pyngrok_config=self.pyngrok_config_v3)
        response2 = agent.get_requests("unknown-tunnel", self.pyngrok_config_v3)

        # THEN
        self.assertEqual(1, len(response1))
        self.assertIsNotNone(1, len(response1[0].uri))
        self.assertIsNotNone(response1[0].id)
        self.assertIsNotNone(response1[0].uri)
        self.assertIsNotNone(response1[0].duration)
        self.assertIsNotNone(response1[0].request)
        self.assertIsNotNone(response1[0].response)
        self.assertEqual(tunnel_name, response1[0].tunnel_name)
        self.assertEqual(0, len(response2))

        # WHEN
        response3 = agent.get_request(response1[0].id, pyngrok_config=self.pyngrok_config_v3)

        # THEN
        self.assertEqual(response1[0].id, response3.id)
        self.assertEqual(tunnel_name, response3.tunnel_name)

        # WHEN
        agent.replay_request(response1[0].id, pyngrok_config=self.pyngrok_config_v3)
        response4 = sorted(agent.get_requests(pyngrok_config=self.pyngrok_config_v3), key=lambda x: x.id)

        # THEN
        self.assertEqual(2, len(response4))
        self.assertEqual(response1[0].id, response4[0].id)
        self.assertEqual(tunnel_name, response4[0].tunnel_name)
        self.assertEqual(tunnel_name, response4[1].tunnel_name)

        # WHEN
        agent.delete_requests(pyngrok_config=self.pyngrok_config_v3)
        response5 = agent.get_requests(pyngrok_config=self.pyngrok_config_v3)

        self.assertEqual(0, len(response5))

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_get_agent_status(self):
        # GIVEN
        ngrok.get_ngrok_process(pyngrok_config=self.pyngrok_config_v3)
        time.sleep(1)

        response = agent.get_agent_status(self.pyngrok_config_v3)

        # THEN
        self.assertEqual("online", response.status)
        self.assertIsNotNone(response.agent_version)
        self.assertIsNotNone(response.uri)



__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import os
from typing import Callable, Optional

from pyngrok.installer import get_ngrok_bin, get_default_ngrok_dir
from pyngrok.log import NgrokLog

DEFAULT_CONFIG_PATH: Optional[str] = None

DEFAULT_NGROK_DIR = get_default_ngrok_dir()
DEFAULT_NGROK_CONFIG_PATH = os.path.join(DEFAULT_NGROK_DIR, "ngrok.yml")
DEFAULT_NGROK_PATH = os.path.join(DEFAULT_NGROK_DIR, get_ngrok_bin())


class PyngrokConfig:
    """
    An object containing ``pyngrok``'s configuration for interacting with the ``ngrok`` binary. All values are
    optional when it is instantiated, and default values will be used for parameters not passed.

    Use :func:`~pyngrok.conf.get_default` and :func:`~pyngrok.conf.set_default` to interact with the default
    ``pyngrok_config``, or pass another instance of this object as the ``pyngrok_config`` keyword arg to most
    methods in the :mod:`~pyngrok.ngrok` module to override the default.

    .. code-block:: python

        from pyngrok import conf, ngrok

        # Here you update the entire default config
        pyngrok_config = conf.PyngrokConfig(ngrok_path="/usr/local/bin/ngrok")
        conf.set_default(pyngrok_config)

        # Here you update just one variable in the default config
        conf.get_default().ngrok_path = "/usr/local/bin/ngrok"

        # Here you leave the default config as-is and pass an override
        pyngrok_config = conf.PyngrokConfig(ngrok_path="/usr/local/bin/ngrok")
        ngrok.connect(pyngrok_config=pyngrok_config)
    """

    def __init__(self,
                 ngrok_path: Optional[str] = None,
                 config_path: Optional[str] = None,
                 auth_token: Optional[str] = None,
                 region: Optional[str] = None,
                 monitor_thread: bool = True,
                 log_event_callback: Optional[Callable[[NgrokLog], None]] = None,
                 startup_timeout: int = 15,
                 max_logs: int = 100,
                 request_timeout: float = 4,
                 start_new_session: bool = False,
                 ngrok_version: str = "v3",
                 api_key: Optional[str] = None,
                 config_version: str = "2") -> None:
        #: The path to the ``ngrok`` binary, defaults to being placed in the same directory as
        #: `ngrok's configs <https://ngrok.com/docs/agent/config/v2>`_.
        self.ngrok_path: str = DEFAULT_NGROK_PATH if ngrok_path is None else ngrok_path
        #: The path to the ``ngrok`` config, defaults to ``None`` and ``ngrok`` manages it.
        self.config_path: Optional[str] = DEFAULT_CONFIG_PATH if config_path is None else config_path
        #: A ``ngrok`` authtoken to pass to commands (overrides what is in the config). If a value is not passed, will
        #: attempt to use the environment variable ``NGROK_AUTHTOKEN`` if it is set.
        self.auth_token: Optional[str] = auth_token or os.environ.get("NGROK_AUTHTOKEN")
        #: The region in which ``ngrok`` should start.
        self.region: Optional[str] = region
        #: Whether ``ngrok`` should continue to be monitored (for logs, etc.) after startup is complete.
        self.monitor_thread: bool = monitor_thread
        #: A callback that will be invoked each time ``ngrok`` emits a log. The function should take
        #: one argument of type :py:class:`str`. ``monitor_thread`` must be set to ``True`` or the function will
        #: stop being called after ``ngrok`` finishes starting.
        self.log_event_callback: Optional[Callable[[NgrokLog], None]] = log_event_callback
        #: The max timeout, in seconds, to wait for ``ngrok`` to start.
        self.startup_timeout: int = startup_timeout
        #: The max number of logs to store in :class:`~pyngrok.process.NgrokProcess`'s ``logs`` variable.
        self.max_logs: int = max_logs
        #: The max timeout, in seconds, when making requests to ``ngrok``'s API.
        self.request_timeout: float = request_timeout
        #: Passed to :py:class:`subprocess.Popen` when launching ``ngrok``. (Python 3 and POSIX only).
        self.start_new_session: bool = start_new_session
        #: The major version of ``ngrok`` installed.
        self.ngrok_version: str = ngrok_version
        #: A ``ngrok`` API key.
        self.api_key: Optional[str] = api_key or os.environ.get("NGROK_API_KEY")
        #: The ``ngrok`` config version.
        self.config_version = config_version


_default_pyngrok_config: PyngrokConfig = PyngrokConfig()


def get_default() -> PyngrokConfig:
    """
    Get the default config to be used with methods in the :mod:`~pyngrok.ngrok` module. To override the
    default individually, the ``pyngrok_config`` keyword arg can also be passed to most of these methods,
    or set a new default config with :func:`~pyngrok.conf.set_default`.

    :return: The default ``pyngrok_config``.
    """
    if _default_pyngrok_config is None:
        set_default(PyngrokConfig())

    return _default_pyngrok_config


def set_default(pyngrok_config: PyngrokConfig) -> None:
    """
    Set a new default config to be used with methods in the :mod:`~pyngrok.ngrok` module. To override the
    default individually, the ``pyngrok_config`` keyword arg can also be passed to most of these methods.

    :param pyngrok_config: The new ``pyngrok_config`` to be used by default.
    """
    global _default_pyngrok_config

    _default_pyngrok_config = pyngrok_config


def get_config_path(pyngrok_config: PyngrokConfig) -> str:
    """
    Return the ``config_path`` if set on the given ``pyngrok_configg``, otherwise return ``ngrok``'s default path.

    :param pyngrok_config: The ``pyngrok`` configuration to first check for a ``config_path``.
    :return: The path to the config file.
    """
    if pyngrok_config.config_path is not None:
        return pyngrok_config.config_path
    else:
        return DEFAULT_NGROK_CONFIG_PATH



#!/usr/bin/env python

__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import logging
from typing import Any, Dict, List, Optional

from pyngrok import conf
from pyngrok.conf import PyngrokConfig
from pyngrok.ngrok import get_ngrok_process, api_request

logger = logging.getLogger(__name__)


class NgrokAgent:
    """
    An object containing information about a ``ngrok`` agent.
    """

    def __init__(self,
                 data: Dict[str, Any]) -> None:
        #: The original tunnel data.
        self.data: Dict[str, Any] = data

        #: The status of the agent.
        self.status: Optional[str] = data.get("status")
        #: The version of the agent.
        self.agent_version: Optional[str] = data.get("agent_version")
        #: The session details for the agent.
        self.session: Any = data.get("session")
        #: The URI of the agent.
        self.uri: Optional[str] = data.get("uri")

    def __repr__(self) -> str:
        return f"<NgrokAgent: \"{self.uri}\">"

    def __str__(self) -> str:  # pragma: no cover
        return f"NgrokAgent: \"{self.uri}\""


class CapturedRequest:
    """
    An object containing a Captured Request from a ``ngrok`` tunnel.
    """

    def __init__(self,
                 data: Dict[str, Any]) -> None:
        #: The original tunnel data.
        self.data: Dict[str, Any] = data

        #: The URI of the captured request.
        self.id: Optional[str] = data.get("id")
        self.uri: Optional[str] = data.get("uri")
        self.tunnel_name: Optional[str] = data.get("tunnel_name")
        self.remote_addr: Optional[str] = data.get("remote_addr")
        self.start: Optional[str] = data.get("start")
        self.duration: Optional[int] = data.get("duration")
        self.request: Any = data.get("request")
        self.response: Any = data.get("response")

    def __repr__(self) -> str:
        return f"<CapturedRequest: \"{self.id}\">"

    def __str__(self) -> str:  # pragma: no cover
        return f"CapturedRequest: \"{self.id}\""


def get_agent_status(pyngrok_config: Optional[PyngrokConfig] = None, ) -> NgrokAgent:
    """
    Get the ``ngrok`` agent status.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The requests made to the tunnels.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    api_url = get_ngrok_process(pyngrok_config).api_url

    return NgrokAgent(api_request(f"{api_url}/api/status", "GET",
                                  timeout=pyngrok_config.request_timeout))


def get_requests(tunnel_name: Optional[str] = None,
                 pyngrok_config: Optional[PyngrokConfig] = None, ) -> List[CapturedRequest]:
    """
    Get the list of requests made to either all tunnels, or the given tunnel name.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param tunnel_name: The tunnel name to filter by.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The requests made to the tunnels.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()
    params = {"tunnel_name": tunnel_name} if tunnel_name else None

    api_url = get_ngrok_process(pyngrok_config).api_url

    requests = []
    for request in api_request(f"{api_url}/api/requests/http", "GET",
                               params=params,
                               timeout=pyngrok_config.request_timeout)["requests"]:
        requests.append(CapturedRequest(request))
    return requests


def get_request(request_id: str,
                pyngrok_config: Optional[PyngrokConfig] = None, ) -> CapturedRequest:
    """
    Get the given request made.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param request_id: The ID of the request to fetch.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    :return: The request made to the tunnel.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    api_url = get_ngrok_process(pyngrok_config).api_url

    return CapturedRequest(api_request(f"{api_url}/api/requests/http/{request_id}", "GET",
                                       timeout=pyngrok_config.request_timeout))


def replay_request(request_id: str,
                   tunnel_name: Optional[str] = None,
                   pyngrok_config: Optional[PyngrokConfig] = None, ) -> None:
    """
    Replay a given request through its original tunnel, or through a different given tunnel.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param request_id: The request ID.
    :param tunnel_name: The name of tunnel to replay the request through.
    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    api_url = get_ngrok_process(pyngrok_config).api_url

    api_request(f"{api_url}/api/requests/http", "POST",
                data={"id": request_id, "tunnel_name": tunnel_name},
                timeout=pyngrok_config.request_timeout)


def delete_requests(pyngrok_config: Optional[PyngrokConfig] = None) -> None:
    """
    Delete request history.

    If ``ngrok`` is not installed at :class:`~pyngrok.conf.PyngrokConfig`'s ``ngrok_path``, calling this method
    will first download and install ``ngrok``.

    If ``ngrok`` is not running, calling this method will first start a process with
    :class:`~pyngrok.conf.PyngrokConfig`.

    :param pyngrok_config: A ``pyngrok`` configuration to use when interacting with the ``ngrok`` binary,
        overriding :func:`~pyngrok.conf.get_default()`.
    """
    if pyngrok_config is None:
        pyngrok_config = conf.get_default()

    api_url = get_ngrok_process(pyngrok_config).api_url

    api_request(f"{api_url}/api/requests/http", "DELETE",
                timeout=pyngrok_config.request_timeout)



__copyright__ = "Copyright (c) 2018-2025 Alex Laird"
__license__ = "MIT"

import os
from unittest import mock

from pyngrok.conf import PyngrokConfig
from tests.testcase import NgrokTestCase


class TestConf(NgrokTestCase):
    def test_auth_token_set_from_env(self):
        # GIVEN
        ngrok_auth_token = "some-auth-token"

        # WHEN
        with mock.patch.dict(os.environ, {"NGROK_AUTHTOKEN": ngrok_auth_token}):
            pyngrok_config = PyngrokConfig()

        # THEN
        self.assertEqual(ngrok_auth_token, pyngrok_config.auth_token)

    def test_api_key_set_from_env(self):
        # GIVEN
        ngrok_api_key = "some-api-key"

        # WHEN
        with mock.patch.dict(os.environ, {"NGROK_API_KEY": ngrok_api_key}):
            pyngrok_config = PyngrokConfig()

        # THEN
        self.assertEqual(ngrok_api_key, pyngrok_config.api_key)
